"""A modification workflow that updates the :term:`LAG` interfaces that are part of an existing IP trunk."""

import ipaddress
import json
from uuid import UUID, uuid4

from orchestrator.forms import FormPage, ReadOnlyField
from orchestrator.forms.validators import UniqueConstrainedList
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, UUIDstr
from orchestrator.utils.json import json_dumps
from orchestrator.workflow import StepList, conditional, done, init, step, workflow
from orchestrator.workflows.steps import resync, store_process_subscription, unsync
from orchestrator.workflows.utils import wrap_modify_initial_input_form
from pydantic import validator
from pydantic_forms.validators import Label

from gso.products.product_blocks.iptrunk import (
    IptrunkInterfaceBlock,
    IptrunkSideBlock,
    IptrunkType,
    PhyPortCapacity,
)
from gso.products.product_blocks.router import RouterVendor
from gso.products.product_types.iptrunk import Iptrunk
from gso.services.netbox_client import NetboxClient
from gso.services.provisioning_proxy import execute_playbook, pp_interaction
from gso.utils.helpers import (
    LAGMember,
    available_interfaces_choices,
    available_interfaces_choices_including_current_members,
    get_router_vendor,
    validate_interface_name_list,
    validate_iptrunk_unique_interface,
)


def initialize_ae_members(subscription: Iptrunk, initial_user_input: dict, side_index: int) -> type[LAGMember]:
    """Initialize the list of AE members."""
    router = subscription.iptrunk.iptrunk_sides[side_index].iptrunk_side_node
    router_vendor = get_router_vendor(router.owner_subscription_id)
    iptrunk_minimum_link = initial_user_input["iptrunk_minimum_links"]
    if router_vendor == RouterVendor.NOKIA:
        iptrunk_speed = initial_user_input["iptrunk_speed"]

        class NokiaLAGMember(LAGMember):
            interface_name: (  # type: ignore[valid-type]
                available_interfaces_choices_including_current_members(
                    router.owner_subscription_id,
                    iptrunk_speed,
                    subscription.iptrunk.iptrunk_sides[side_index].iptrunk_side_ae_members,
                )
                if iptrunk_speed == subscription.iptrunk.iptrunk_speed
                else (
                    available_interfaces_choices(
                        router.owner_subscription_id,
                        initial_user_input["iptrunk_speed"],
                    )
                )
            )

        class NokiaAeMembers(UniqueConstrainedList[NokiaLAGMember]):
            min_items = iptrunk_minimum_link

        ae_members = NokiaAeMembers
    else:

        class JuniperAeMembers(UniqueConstrainedList[LAGMember]):
            min_items = iptrunk_minimum_link

        ae_members = JuniperAeMembers  # type: ignore[assignment]
    return ae_members  # type: ignore[return-value]


def initial_input_form_generator(subscription_id: UUIDstr) -> FormGenerator:
    """Gather input from the operator on the interfaces that should be modified."""
    subscription = Iptrunk.from_subscription(subscription_id)

    class ModifyIptrunkForm(FormPage):
        tt_number: str
        geant_s_sid: str = subscription.iptrunk.geant_s_sid
        iptrunk_description: str = subscription.iptrunk.iptrunk_description
        iptrunk_type: IptrunkType = subscription.iptrunk.iptrunk_type
        warning_label: Label = (
            "Changing the PhyPortCapacity will result in the deletion of all AE members. "
            "You will need to add the new AE members in the next steps."  # type: ignore[assignment]
        )
        iptrunk_speed: PhyPortCapacity = subscription.iptrunk.iptrunk_speed
        iptrunk_minimum_links: int = subscription.iptrunk.iptrunk_minimum_links
        iptrunk_isis_metric: int = ReadOnlyField(subscription.iptrunk.iptrunk_isis_metric)
        iptrunk_ipv4_network: ipaddress.IPv4Network = ReadOnlyField(subscription.iptrunk.iptrunk_ipv4_network)
        iptrunk_ipv6_network: ipaddress.IPv6Network = ReadOnlyField(subscription.iptrunk.iptrunk_ipv6_network)

    initial_user_input = yield ModifyIptrunkForm
    ae_members_side_a = initialize_ae_members(subscription, initial_user_input.dict(), 0)

    class ModifyIptrunkSideAForm(FormPage):
        class Config:
            title = "Provide subscription details for side A of the trunk."

        side_a_node: str = ReadOnlyField(subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.router_fqdn)
        side_a_ae_iface: str = ReadOnlyField(subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_iface)
        side_a_ae_geant_a_sid: str = subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_geant_a_sid
        side_a_ae_members: ae_members_side_a = (  # type: ignore[valid-type]
            subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_members
            if initial_user_input.iptrunk_speed == subscription.iptrunk.iptrunk_speed
            else []
        )

        @validator("side_a_ae_members", allow_reuse=True)
        def validate_iptrunk_unique_interface_side_a(cls, side_a_ae_members: list[LAGMember]) -> list[LAGMember]:
            return validate_iptrunk_unique_interface(side_a_ae_members)

        @validator("side_a_ae_members", allow_reuse=True)
        def validate_interface_name_members(cls, side_a_ae_members: list[LAGMember]) -> list[LAGMember]:
            vendor = subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.vendor
            return validate_interface_name_list(side_a_ae_members, vendor)

    user_input_side_a = yield ModifyIptrunkSideAForm
    ae_members_side_b = initialize_ae_members(subscription, initial_user_input.dict(), 1)

    class ModifyIptrunkSideBForm(FormPage):
        class Config:
            title = "Provide subscription details for side B of the trunk."

        side_b_node: str = ReadOnlyField(subscription.iptrunk.iptrunk_sides[1].iptrunk_side_node.router_fqdn)
        side_b_ae_iface: str = ReadOnlyField(subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_iface)
        side_b_ae_geant_a_sid: str = subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_geant_a_sid
        side_b_ae_members: ae_members_side_b = (  # type: ignore[valid-type]
            subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_members
            if initial_user_input.iptrunk_speed == subscription.iptrunk.iptrunk_speed
            else []
        )

        @validator("side_b_ae_members", allow_reuse=True)
        def validate_iptrunk_unique_interface_side_b(cls, side_b_ae_members: list[LAGMember]) -> list[LAGMember]:
            return validate_iptrunk_unique_interface(side_b_ae_members)

        @validator("side_b_ae_members", allow_reuse=True)
        def validate_interface_name_members(cls, side_b_ae_members: list[LAGMember]) -> list[LAGMember]:
            vendor = subscription.iptrunk.iptrunk_sides[1].iptrunk_side_node.vendor
            return validate_interface_name_list(side_b_ae_members, vendor)

    user_input_side_b = yield ModifyIptrunkSideBForm

    return initial_user_input.dict() | user_input_side_a.dict() | user_input_side_b.dict()


@step("Update subscription")
def modify_iptrunk_subscription(
    subscription: Iptrunk,
    geant_s_sid: str,
    iptrunk_type: IptrunkType,
    iptrunk_description: str,
    iptrunk_speed: PhyPortCapacity,
    iptrunk_minimum_links: int,
    side_a_ae_geant_a_sid: str,
    side_a_ae_members: list[dict],
    side_b_ae_geant_a_sid: str,
    side_b_ae_members: list[dict],
) -> State:
    """Modify the subscription in the service database, reflecting the changes to the newly selected interfaces."""
    # Prepare the list of removed AE members
    previous_ae_members = [
        [
            {
                "interface_name": member.interface_name,
                "interface_description": member.interface_description,
            }
            for member in side.iptrunk_side_ae_members
        ]
        for side in subscription.iptrunk.iptrunk_sides
    ]
    removed_ae_members = []

    for side_index in range(2):
        previous_members = previous_ae_members[side_index]
        current_members = side_a_ae_members if side_index == 0 else side_b_ae_members
        removed_ae_members.append([ae_member for ae_member in previous_members if ae_member not in current_members])

    subscription.iptrunk.geant_s_sid = geant_s_sid
    subscription.iptrunk.iptrunk_description = iptrunk_description
    subscription.iptrunk.iptrunk_type = iptrunk_type
    subscription.iptrunk.iptrunk_speed = iptrunk_speed
    subscription.iptrunk.iptrunk_minimum_links = iptrunk_minimum_links

    subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_geant_a_sid = side_a_ae_geant_a_sid
    #  Flush the old list of member interfaces
    subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_members.clear()
    #  And update the list to only include the new member interfaces
    for member in side_a_ae_members:
        subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_members.append(
            IptrunkInterfaceBlock.new(subscription_id=uuid4(), **member),
        )

    subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_geant_a_sid = side_b_ae_geant_a_sid
    subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_members.clear()
    for member in side_b_ae_members:
        subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_members.append(
            IptrunkInterfaceBlock.new(subscription_id=uuid4(), **member),
        )

    subscription.description = f"IP trunk, geant_s_sid:{geant_s_sid}"

    return {
        "subscription": subscription,
        "removed_ae_members": removed_ae_members,
        "previous_ae_members": previous_ae_members,
    }


@step("Provision IP trunk interface [DRY RUN]")
def provision_ip_trunk_iface_dry(
    subscription: Iptrunk,
    process_id: UUIDstr,
    callback_route: str,
    tt_number: str,
    removed_ae_members: list[str],
) -> State:
    """Perform a dry run of deploying the updated IP trunk."""
    extra_vars = {
        "wfo_trunk_json": json.loads(json_dumps(subscription)),
        "dry_run": True,
        "verb": "deploy",
        "config_object": "trunk_interface",
        "removed_ae_members": removed_ae_members,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Deploy config for "
        f"{subscription.iptrunk.geant_s_sid} ",
    }

    execute_playbook(
        playbook_name="iptrunks.yaml",
        callback_route=callback_route,
        inventory=f"{subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.router_fqdn}\n"
        f"{subscription.iptrunk.iptrunk_sides[1].iptrunk_side_node.router_fqdn}\n",
        extra_vars=extra_vars,
    )

    return {"subscription": subscription}


@step("Provision IP trunk interface [FOR REAL]")
def provision_ip_trunk_iface_real(
    subscription: Iptrunk,
    process_id: UUIDstr,
    callback_route: str,
    tt_number: str,
    removed_ae_members: list[str],
) -> State:
    """Provision the new IP trunk with updated interfaces."""
    extra_vars = {
        "wfo_trunk_json": json.loads(json_dumps(subscription)),
        "dry_run": False,
        "verb": "deploy",
        "config_object": "trunk_interface",
        "removed_ae_members": removed_ae_members,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Deploy config for "
        f"{subscription.iptrunk.geant_s_sid} ",
    }

    execute_playbook(
        playbook_name="iptrunks.yaml",
        callback_route=callback_route,
        inventory=f"{subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.router_fqdn}\n"
        f"{subscription.iptrunk.iptrunk_sides[1].iptrunk_side_node.router_fqdn}\n",
        extra_vars=extra_vars,
    )

    return {"subscription": subscription}


def _netbox_update_interfaces(
    subscription_id: UUID,
    side_block: IptrunkSideBlock,
    removed_ae_members: list[dict],
    previous_ae_members: list[dict],
) -> None:
    nbclient = NetboxClient()

    # Free removed interfaces
    for member in removed_ae_members:
        nbclient.free_interface(side_block.iptrunk_side_node.router_fqdn, member["interface_name"])

    # Attach physical interfaces to :term:`LAG`
    # Update interface description to subscription ID
    # Reserve interfaces
    for interface in side_block.iptrunk_side_ae_members:
        if any(ae_member["interface_name"] == interface.interface_name for ae_member in previous_ae_members):
            continue
        if side_block.iptrunk_side_ae_iface:
            nbclient.attach_interface_to_lag(
                device_name=side_block.iptrunk_side_node.router_fqdn,
                lag_name=side_block.iptrunk_side_ae_iface,
                iface_name=interface.interface_name,
                description=str(subscription_id),
            )

        nbclient.reserve_interface(
            device_name=side_block.iptrunk_side_node.router_fqdn,
            iface_name=interface.interface_name,
        )


@step("Netbox: Reserve side A interfaces")
def netbox_update_interfaces_side_a(
    subscription: Iptrunk, removed_ae_members: list[list[dict]], previous_ae_members: list[list[dict]]
) -> None:
    """Update Netbox such that it contains the new interfaces on side A."""
    _netbox_update_interfaces(
        subscription.subscription_id,
        subscription.iptrunk.iptrunk_sides[0],
        removed_ae_members[0],
        previous_ae_members[0],
    )


@step("Netbox: Reserve side B interfaces")
def netbox_update_interfaces_side_b(
    subscription: Iptrunk, removed_ae_members: list[list[dict]], previous_ae_members: list[list[dict]]
) -> None:
    """Update Netbox such that it contains the new interfaces on side B."""
    _netbox_update_interfaces(
        subscription.subscription_id,
        subscription.iptrunk.iptrunk_sides[1],
        removed_ae_members[1],
        previous_ae_members[1],
    )


def _netbox_allocate_interfaces(side_block: IptrunkSideBlock, previous_ae_members: list[dict]) -> None:
    nbclient = NetboxClient()

    for interface in side_block.iptrunk_side_ae_members:
        if any(ae_member["interface_name"] == interface.interface_name for ae_member in previous_ae_members):
            continue
        nbclient.allocate_interface(
            device_name=side_block.iptrunk_side_node.router_fqdn,
            iface_name=interface.interface_name,
        )

    # detach the old interfaces from lag
    if side_block.iptrunk_side_ae_iface:
        nbclient.detach_interfaces_from_lag(
            device_name=side_block.iptrunk_side_node.router_fqdn,
            lag_name=side_block.iptrunk_side_ae_iface,
        )


@step("Netbox: Allocate side A interfaces")
def allocate_interfaces_in_netbox_side_a(subscription: Iptrunk, previous_ae_members: list[list[dict]]) -> None:
    """Allocate the :term:`LAG` interfaces on side A in Netbox.

    Attach the :term:`LAG` interface to the physical interface detach old one from the :term:`LAG`.
    """
    _netbox_allocate_interfaces(subscription.iptrunk.iptrunk_sides[0], previous_ae_members[0])


@step("Netbox: Allocate side B interfaces")
def allocate_interfaces_in_netbox_side_b(subscription: Iptrunk, previous_ae_members: list[list[dict]]) -> None:
    """Allocate the :term:`LAG` interface on side B in Netbox.

    Attach the :term:`LAG` interface to the physical interface detach old one from the :term:`LAG`.
    """
    _netbox_allocate_interfaces(subscription.iptrunk.iptrunk_sides[1], previous_ae_members[1])


@workflow(
    "Modify IP Trunk interface",
    initial_input_form=wrap_modify_initial_input_form(initial_input_form_generator),
    target=Target.MODIFY,
)
def modify_trunk_interface() -> StepList:
    """Modify the interfaces that are part of an IP trunk.

    * Update the subscription in the database
    * Reserve new interfaces in Netbox
    * Provision the updated version of the IP trunk, first as a dry run
    * Allocate the reserved interfaces in Netbox
    """
    side_a_is_nokia = conditional(
        lambda state: get_router_vendor(
            state["subscription"]["iptrunk"]["iptrunk_sides"][0]["iptrunk_side_node"]["owner_subscription_id"]
        )
        == RouterVendor.NOKIA
    )
    side_b_is_nokia = conditional(
        lambda state: get_router_vendor(
            state["subscription"]["iptrunk"]["iptrunk_sides"][1]["iptrunk_side_node"]["owner_subscription_id"]
        )
        == RouterVendor.NOKIA
    )
    return (
        init
        >> store_process_subscription(Target.MODIFY)
        >> unsync
        >> modify_iptrunk_subscription
        >> side_a_is_nokia(netbox_update_interfaces_side_a)
        >> side_b_is_nokia(netbox_update_interfaces_side_b)
        >> pp_interaction(provision_ip_trunk_iface_dry)
        >> pp_interaction(provision_ip_trunk_iface_real)
        >> side_a_is_nokia(allocate_interfaces_in_netbox_side_a)
        >> side_b_is_nokia(allocate_interfaces_in_netbox_side_b)
        >> resync
        >> done
    )
